import httpx
import asyncio
import logging
from io import StringIO
from tenacity import retry
from tenacity.stop import stop_after_attempt
from . import parser as gp
from asyncio import Semaphore

lock = Semaphore(10)

def jsonify(txt):
    entity = []
    accession = None
    in_table = False
    for line in StringIO(txt):
        try:
            if (isinstance(line, bytes)):
                line = line.decode()
        except:
            pass
        line = line.strip()

        # ignore table details for now

        if (line.endswith('table_begin')):
            in_table = True
        if (line.endswith('table_end')):
            in_table = False
            continue
        if (in_table):
            continue

        # ignore comments

        if (line.startswith('#')):
            continue

        if (line.startswith('^')):
            if (accession is not None):
                yield (gp._parse_single_entity_soft(entity))
            accession = gp._split_on_first_equal(line)[1]
            entity = []

        entity.append(line)
    yield (gp._parse_single_entity_soft(entity))


async def get_geo_accession_soft(accession: str, targ: str='all', view: str= "brief"):
    logging.info(f'Accessing accession {accession} in SOFT format')
    url = (
        "https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?targ={}&acc={}&form=text&view={}"
        .format(targ, accession, view))
    
    @retry(stop=stop_after_attempt(10))
    async def _perform_get(url: str) -> str:
        logging.info(f"fetching{url}")
        async with httpx.AsyncClient() as client:
            resp = await client.get(url)
            resp.raise_for_status()
        return resp.text

    res = await _perform_get(url)
    for i in jsonify(res):
        yield(i.json())


async def stuff(acc):
    print('here')
    async with lock:
        async for i in get_geo_accession_soft(acc):
            print(i)

async def abc():
    results = asyncio.gather(*[stuff(acc) for acc in gp.get_geo_accessions(etyp='GSE',batch_size=10)])

if __name__ == '__main__':
    asyncio.run(abc())


class BaseType(object):
    def __init__(self, name, nullable):
        self.name = name
        self.nullable = nullable

    def __repr__(self):
        return "<{} class name:{}>".format(self.__type__, self.name)


################################################
#
#
#  Classes for simple entities:
#
#  Strings, Dates(timestamp), Boolean,
#  Long, Integer
#
#
################################################


class StringType(BaseType):
    __type__ = "StringType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return {
            "fields": {"keyword": {"ignore_above": 256, "type": "keyword"}},
            "type": "text",
        }


class DateType(BaseType):
    __type__ = "DateType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return {"type": "timestamp"}


class BooleanType(BaseType):
    __type__ = "BooleanType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return {"type": "boolean"}


class LongType(BaseType):
    __type__ = "LongType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return {"type": "long"}


class IntType(BaseType):
    __type__ = "IntType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)

    def mapping(self):
        return {"type": "integer"}


################################################
#
#
#  Classes for complex entities:
#
#    Objects
#    Nested (arrays)
#
#
################################################


class ObjectType(BaseType):
    # spark type: struct
    # ES type:    object
    __type__ = "ObjectType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)
        self.fields = {}

    def mapping(self):
        return {
            "properties": dict(
                (k, self.fields[k].mapping()) for k in self.fields.keys()
            )
        }


class NestedType(BaseType):
    # spark type: array
    # ES type:    nested
    __type__ = "NestedType"

    def __init__(self, name, nullable):
        super().__init__(name, nullable)
        self.fields = {}

    def mapping(self):
        return {
            "include_in_parent": True,
            "type": "nested",
            "properties": dict(
                (k, self.fields[k].mapping()) for k in self.fields.keys()
            ),
        }


def walk_schema(schema, parent=None):
    """Walk a pyspark "schema.jsonValue" schema

    This function converts a pyspark schema into
    a set of classes that can then convert the schema
    to an elasticsearch mapping.

    Parameters
    ----------
    schema: a dict
        Typically, this would come from `df.schema.jsonValue()`. It could
        also be loaded from a string or file using the `json` library.
    parent: a container class, so an `ObjectType` or a `NestedType`
        If empty, will be initialized with an `ObjectType` with
        name = "root".

    >>> s = walk_schema(df.schema.jsonValue(), ObjectType(name='root'))
    >>> s.mapping()
    """

    if parent is None:
        parent = ObjectType(name="root")

    for field in schema["fields"]:
        if field["type"] == "string":
            parent.fields[field["name"]] = StringType(
                name=field["name"], nullable=field["nullable"]
            )
        if field["type"] == "integer":
            parent.fields[field["name"]] = IntType(
                name=field["name"], nullable=field["nullable"]
            )
        if field["type"] == "long":
            parent.fields[field["name"]] = LongType(
                name=field["name"], nullable=field["nullable"]
            )
        if field["type"] == "timestamp":
            parent.fields[field["name"]] = DateType(
                name=field["name"], nullable=field["nullable"]
            )
        if field["type"] == "boolean":
            parent.fields[field["name"]] = BooleanType(
                name=field["name"], nullable=field["nullable"]
            )
        if type(field["type"]) == dict:
            n_field = field["type"]
            if n_field["type"] == "struct":
                newp = ObjectType(field["name"], field["nullable"])
                parent.fields[field["name"]] = walk_schema(n_field, parent=newp)

            if n_field["type"] == "array":
                newp = NestedType(field["name"], field["nullable"])
                parent.fields[field["name"]] = walk_schema(
                    n_field["elementType"], parent=newp
                )
    return parent



"""Tooling and utilities for working with ontologies

```
from omicidx.ontologies.utils import parse_ontology

cl = ontology_from_obo_library('cl.obo')
cl_item = cl['CL:0000900']
# lowercase definition
cl_item.definition.lower()
# term
cl_item.name
# all direct relationships
cl_item.relationships
# superclasses (all ancestors)
#  and an equivalent subclasses is available (all descendents)
list(cl_item.superclasses())
"""
from pronto import Ontology
import requests

ONTOLOGY_CATALOG_URL = "http://obofoundry.org/registry/ontologies.jsonld"


def ontology_from_obo_library(ontology_short_name: str) -> Ontology:
    """parse an ontology

    This is taken directly from pronto Ontology right
    now. Parse an OBO, JSON-graph, or OWL format ontology.

    Parameters
    ==========
    ontology_short_name: str
        The short name from (cl.obo for cell line, ncit.obo for NCIT, etc.)

    Return
    ======
    An pronto Ontology object
    """
    ont = Ontology.from_obo_library(ontology_short_name)
    return ont


def get_all_ontologies_from_obo_library():
    """get complete list of ontologies from ontobee

    Return
    ======
    list of dict, each representing an ontology.

    ```
    ont_ids = list([ont['id'] for ont in get_all_ontologies_from_obo_library()])
    ont_ids[0:5]
    ```

    """
    resp = requests.get(ONTOLOGY_CATALOG_URL)
    return resp.json()["ontologies"]
